# pyfoam\pyfoam\core\symbolicly\symbolic_regression.py

import numpy as np
import sympy as sp
from itertools import product
from typing import List, Tuple, Dict, Any, Optional

# =============================================================================
# Helper function: generate_multi_indices
# =============================================================================
def generate_multi_indices(n_vars: int, max_degree: int) -> List[Tuple[int, ...]]:
    """
    Generate all multi-indices (tuples of nonnegative integers) for a polynomial in n_vars
    variables with a total degree less than or equal to max_degree.
    
    Args:
        n_vars (int): Number of variables.
        max_degree (int): Maximum total degree of the polynomial.
        
    Returns:
        List[Tuple[int, ...]]: A list of tuples. Each tuple represents the exponents 
                               of a term in the polynomial basis.
                               
    Example:
        generate_multi_indices(2, 2) returns:
        [(0, 0), (1, 0), (0, 1), (2, 0), (1, 1), (0, 2)]
    """
    indices = []
    for total_degree in range(max_degree + 1):
        for exponents in product(range(total_degree + 1), repeat=n_vars):
            if sum(exponents) == total_degree:
                indices.append(exponents)
    return indices

# =============================================================================
# Helper function: build_design_matrix
# =============================================================================
def build_design_matrix(points: np.ndarray, poly_indices: List[Tuple[int, ...]]) -> np.ndarray:
    """
    Build the design (feature) matrix for polynomial regression.
    
    Each column of the design matrix corresponds to one term in the polynomial basis, 
    evaluated at all given points.
    
    Args:
        points (np.ndarray): Array of shape (N, n_vars) where N is the number of points.
        poly_indices (List[Tuple[int, ...]]): List of tuples representing the exponent for each term.
        
    Returns:
        np.ndarray: Design matrix of shape (N, number_of_terms)
    """
    N, n_vars = points.shape
    n_terms = len(poly_indices)
    A = np.ones((N, n_terms))
    for j, exponents in enumerate(poly_indices):
        for i in range(n_vars):
            A[:, j] *= points[:, i] ** exponents[i]
    return A

# =============================================================================
# Main Function: symbolic_regression_solution
# =============================================================================
def symbolic_regression_solution(mesh: Any,
                                   numerical_values: np.ndarray,
                                   var_names: List[str],
                                   poly_degree: int = 3,
                                   bc_dict: Optional[Dict[int, float]] = None,
                                   ic_dict: Optional[Dict[int, float]] = None) -> sp.Expr:
    """
    Perform symbolic regression to recover an approximate analytical solution from
    numerical data provided on a CFD mesh.
    
    This function constructs a polynomial basis (up to a specified degree) and solves 
    a least-squares problem to determine the coefficients. It also allows for the incorporation 
    of boundary conditions (bc_dict) and initial conditions (ic_dict) by overriding the 
    corresponding numerical values.
    
    Args:
        mesh (Any): A CFD mesh object (e.g., an instance of StructuredMesh or UnstructuredMesh) 
                    that has at least a 'nodes' attribute (numpy array of shape (N, dimensions)) 
                    and a 'bounds' attribute (list of (lower, upper) tuples).
        numerical_values (np.ndarray): Numerical solution values (can be multi-dimensional) 
                                       associated with each mesh node. Will be flattened to a 1D array.
        var_names (List[str]): List of variable names (e.g., ['x', 'y'] or ['x', 't']).
        poly_degree (int, optional): Maximum polynomial degree in the ansatz. Default is 3.
        bc_dict (Optional[Dict[int, float]]): Optional dictionary mapping mesh node indices to 
                                              boundary condition values. These values will override 
                                              the corresponding numerical values.
        ic_dict (Optional[Dict[int, float]]): Optional dictionary mapping mesh node indices to 
                                              initial condition values. These values will override 
                                              the corresponding numerical values.
                                              
    Returns:
        sp.Expr: A sympy expression representing the approximate analytical solution.
    """
    # Retrieve the mesh nodes (assumed to be a NumPy array of shape (N, dimensions))
    points = mesh.nodes  # shape: (N, dimensions)
    N = points.shape[0]
    
    # Flatten the numerical values (if not already a 1D array)
    values = np.array(numerical_values).flatten()
    
    # Override numerical values with prescribed boundary condition values if available.
    if bc_dict is not None:
        for idx, val in bc_dict.items():
            values[idx] = val
            
    # Override numerical values with prescribed initial condition values if available.
    if ic_dict is not None:
        for idx, val in ic_dict.items():
            values[idx] = val

    # Generate the polynomial multi-indices for the given number of variables and degree.
    poly_indices = generate_multi_indices(len(var_names), poly_degree)
    
    # Build the design matrix using the mesh nodes.
    A = build_design_matrix(points, poly_indices)
    
    # Solve the least squares problem A * coeffs ≈ values.
    coeffs, residuals, rank, s = np.linalg.lstsq(A, values, rcond=None)
    
    # Construct the symbolic expression by combining the coefficients with the corresponding monomials.
    symbols = sp.symbols(var_names)
    expr = 0
    for coef, exponents in zip(coeffs, poly_indices):
        term = coef
        for sym, power in zip(symbols, exponents):
            term *= sym ** power
        expr += term
        
    # Simplify the symbolic expression.
    expr = sp.simplify(expr)
    return expr

# =============================================================================
# PDE-Specific Approximation Functions
# =============================================================================
def approximate_poisson_solution(mesh: Any,
                                 numerical_values: np.ndarray,
                                 poly_degree: int = 3,
                                 bc: Optional[Any] = None,
                                 ic: Optional[Any] = None) -> sp.Expr:
    """
    Approximate the symbolic solution for the Poisson equation.
    
    Assumes a spatial domain (e.g., 2D: variables 'x' and 'y'). The boundary condition (bc)
    and (if applicable) the initial condition (ic) are applied to override numerical values.
    
    Args:
        mesh (Any): CFD mesh object.
        numerical_values (np.ndarray): Numerical solution values on the mesh nodes.
        poly_degree (int, optional): Maximum polynomial degree. Default is 3.
        bc (Optional[Any]): A BoundaryCondition instance (e.g., DirichletCondition) from which
                            a boundary condition dictionary is obtained.
        ic (Optional[Any]): An InitialCondition instance (e.g., ScalarInitialCondition) for initial values.
    
    Returns:
        sp.Expr: Symbolic approximation for the Poisson equation.
    """
    bc_dict = bc.apply(mesh, t=0) if bc is not None else None
    ic_dict = ic.apply(mesh, t=0) if ic is not None else None
    return symbolic_regression_solution(mesh, numerical_values, ['x', 'y'], poly_degree, bc_dict, ic_dict)

def approximate_helmholtz_solution(mesh: Any,
                                   numerical_values: np.ndarray,
                                   poly_degree: int = 3,
                                   bc: Optional[Any] = None) -> sp.Expr:
    """
    Approximate the symbolic solution for the Helmholtz equation.
    
    Assumes a spatial domain (e.g., 2D: variables 'x' and 'y'). Boundary conditions (bc)
    are incorporated as needed.
    
    Args:
        mesh (Any): CFD mesh object.
        numerical_values (np.ndarray): Numerical solution values on the mesh nodes.
        poly_degree (int, optional): Maximum polynomial degree.
        bc (Optional[Any]): A BoundaryCondition instance providing boundary values.
    
    Returns:
        sp.Expr: Symbolic approximation for the Helmholtz equation.
    """
    bc_dict = bc.apply(mesh, t=0) if bc is not None else None
    return symbolic_regression_solution(mesh, numerical_values, ['x', 'y'], poly_degree, bc_dict)

def approximate_heat_solution(mesh: Any,
                              numerical_values: np.ndarray,
                              poly_degree: int = 3,
                              bc: Optional[Any] = None,
                              ic: Optional[Any] = None) -> sp.Expr:
    """
    Approximate the symbolic solution for the Heat equation.
    
    Assumes a space-time domain (e.g., variables 'x' and 't'). Both boundary (bc) and 
    initial (ic) conditions can be incorporated.
    
    Args:
        mesh (Any): CFD mesh object.
        numerical_values (np.ndarray): Numerical solution values on the mesh nodes.
        poly_degree (int, optional): Maximum polynomial degree.
        bc (Optional[Any]): A BoundaryCondition instance providing boundary values.
        ic (Optional[Any]): An InitialCondition instance providing initial values.
    
    Returns:
        sp.Expr: Symbolic approximation for the Heat equation.
    """
    bc_dict = bc.apply(mesh, t=0) if bc is not None else None
    ic_dict = ic.apply(mesh, t=0) if ic is not None else None
    return symbolic_regression_solution(mesh, numerical_values, ['x', 't'], poly_degree, bc_dict, ic_dict)

def approximate_general_pde_solution(mesh: Any,
                                     numerical_values: np.ndarray,
                                     var_names: List[str],
                                     poly_degree: int = 3,
                                     bc: Optional[Any] = None,
                                     ic: Optional[Any] = None) -> sp.Expr:
    """
    Approximate the symbolic solution for a general PDE.
    
    Args:
        mesh (Any): CFD mesh object.
        numerical_values (np.ndarray): Numerical solution values on the mesh nodes.
        var_names (List[str]): List of variable names (e.g., ['x', 'y', 't']).
        poly_degree (int, optional): Maximum polynomial degree.
        bc (Optional[Any]): A BoundaryCondition instance providing boundary values.
        ic (Optional[Any]): An InitialCondition instance providing initial values.
        
    Returns:
        sp.Expr: Symbolic approximation for the general PDE.
    """
    bc_dict = bc.apply(mesh, t=0) if bc is not None else None
    ic_dict = ic.apply(mesh, t=0) if ic is not None else None
    return symbolic_regression_solution(mesh, numerical_values, var_names, poly_degree, bc_dict, ic_dict)
